/**
 * EasyBeans
 * Copyright (C) 2008 Bull S.A.S.
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: PropertyResolverHelper.java 4389 2008-12-15 13:48:57Z alitokmen $
 * --------------------------------------------------------------------------
 */
package org.ow2.util.xmlconfig.properties;

import java.util.LinkedList;
import java.util.List;

/**
 * Helper to resolve properties.
 * @author Loris Bouzonnet
 */
public class PropertyResolverHelper {

    /**
     * Resolve a property with a given instance of {@link IPropertyResolver}. A
     * property is resolved in two steps: Firstly, the property is parsed to
     * build the corresponding {@link Property}. Secondly, the property is
     * compiled to a {@link String}.
     * @param propertyResolver an instance {@link IPropertyResolver}
     * @param propertyValue a property to resolve
     * @return the resolved property
     * @throws IllegalArgumentException if the given property cannot be resolved
     */
    public static String resolve(final IPropertyResolver propertyResolver, final String propertyValue)
            throws IllegalArgumentException {

        // Build the chain
        Property property = new Property();
        buildChain(property, propertyValue);

        // Resolve the properties
        return resolveProperty(propertyResolver, property);
    }

    /**
     * Parse a given property to resolve and add its result to a given instance
     * of {@link Property}.
     * @param property a property to add the resolved property
     * @param propertyValue a property to resolve
     */
    private static void buildChain(final Property property, final String propertyValue) {

        String tmp = propertyValue;

        // While some properties need to be parsed
        while (!tmp.equals("")) {
            int imin = tmp.indexOf("${");
            String prefix;
            Property subProperty;
            // Only a prefix
            if (imin == -1) {
                prefix = tmp;
                subProperty = null;
                // Only a sub property
            } else if (imin == 0) {
                prefix = "";
                subProperty = new Property();
                // Prefix & sub property
            } else {
                prefix = tmp.substring(0, imin);
                subProperty = new Property();
            }
            // Check for one closing bracket too many
            if (prefix.contains("}")) {
                throw new IllegalArgumentException(tmp + " is not a valid expression: ${ missing");
            }
            if (subProperty != null) {
                int imax = imin + 2;
                int counter = 1;
                // Search for one corresponding closing bracket
                while (counter != 0 && imax < tmp.length()) {
                    if (tmp.charAt(imax) == '}') {
                        counter--;
                    } else if (tmp.charAt(imax) == '{') {
                        counter++;
                    }
                    imax++;
                }
                // Check for one opening bracket too many
                if (counter != 0 && imax == tmp.length()) {
                    throw new IllegalArgumentException(tmp + " is not a valid expression: } missing");
                }
                String sProperty = tmp.substring(imin + 2, imax - 1);
                // Check for empty property
                if (sProperty.equals("")) {
                    throw new IllegalArgumentException("The chain ${} is forbidden.");
                }
                // Build the sub property
                buildChain(subProperty, sProperty);
                tmp = tmp.substring(imax);
            } else {
                tmp = "";
            }
            // Construct a prefixed property with the found prefix and sub
            // property and add it to the given property
            property.getPrefixedProperties().add(new PrefixedProperty(prefix, subProperty));
        }
    }

    /**
     * Compile a given {@link Property} to a {@link String} with a given
     * {@link IPropertyResolver}.
     * @param propertyResolver a property to add the resolved property
     * @param property a property
     * @return a corresponding string
     */
    private static String resolveProperty(final IPropertyResolver propertyResolver, final Property property) {
        String result = "";
        for (PrefixedProperty prefixedProperty : property.getPrefixedProperties()) {
            result += prefixedProperty.getPrefix();
            Property subProperty = prefixedProperty.getProperty();
            if (subProperty != null) {
                String resolvedProperty = propertyResolver.resolve(resolveProperty(propertyResolver, subProperty));
                if (resolvedProperty == null) {
                    throw new IllegalArgumentException("Cannot resolve: " + subProperty);
                }
                result += resolvedProperty;
            }
        }
        return result;
    }

    /**
     * A {@link Property} is a list of {@link PrefixedProperty}.
     */
    private static class Property {

        /**
         * A list of prefixed property.
         */
        private final List<PrefixedProperty> prefixedProperties;

        /**
         * Construct a new property with an empty list of prefixed property.
         */
        public Property() {
            prefixedProperties = new LinkedList<PrefixedProperty>();
        }

        /**
         * @return the list of prefixed property
         */
        public List<PrefixedProperty> getPrefixedProperties() {
            return prefixedProperties;
        }

        @Override
        public String toString() {
            String result = "";
            for (PrefixedProperty prefixedProperty : prefixedProperties) {
                result += prefixedProperty;
            }
            return result;
        }

    }

    /**
     * A {@link PrefixedProperty} is composed of a prefix and a {@link Property}.
     */
    private static class PrefixedProperty {

        /**
         * A prefix.
         */
        private final String prefix;

        /**
         * A property.
         */
        private final Property property;

        /**
         * Construct a new prefixed property with a prefix and a property.
         * @param prefix a prefix
         * @param property a property
         */
        public PrefixedProperty(final String prefix, final Property property) {
            this.prefix = prefix;
            this.property = property;
        }

        /**
         * @return the prefix
         */
        public String getPrefix() {
            return prefix;
        }

        /**
         * @return the property
         */
        public Property getProperty() {
            return property;
        }

        @Override
        public String toString() {
            return prefix + ((property == null) ? "" : ("${" + property + "}"));
        }

    }

}
